# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2005 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.

import threading
import urllib
import string
import os
import socket
import HTTPSURLopener
import OpenSSL
import CommonErrors
import exceptions


class FileNameException(exceptions.Exception): pass

def get_base_filename_from_url(url, legal_exts):
    """ Safely unquotes a URL and gets the base file name from it.
        We're not using urlparse here because it doesn't un-escape stuff """

    unquoted = url
    last_unquoted = None
    count = 5
    
    # Keep unquoting the string until the last two unquote operations
    # produce the same string
    while (unquoted != last_unquoted) and (count > 0):
        last_unquoted = unquoted
        unquoted = urllib.unquote_plus(unquoted)
        count = count - 1

    # If after 5 iterations of unquoting, the strings still aren't the same,
    # something is wrong.
    if (count == 0) and (unquoted != last_unquoted):
        raise FileNameException("URL quoting level too deep.")

    # Try to grab the filename off the end of the URL
    index = url.rfind('/')
    if index is -1:
        raise FileNameException("No separator in URL.")
    filename = url[index+1:]

    # Only accept certain file extensions
    ext_ok = False
    for ext in legal_exts:
        if filename.endswith(ext):
            ext_ok = True
            break

    if not ext_ok:
        raise FileNameException("Extension was not allowed.")

    # FIXME: what other validation can we do here?
    safe_list = ['_', '-', '.', '+']
    for c in filename:
        # For now, legal characters are '_-.' plus alphanumeric
        if c in safe_list or c.isalnum():
            pass
        else:
            raise FileNameException("Illegal character '%s' encountered." % c)

    return filename


class FileDownloader(threading.Thread):

    def __init__(self, callback, cb_data, url, target_dir, legal_exts, certs):
        self._callback = callback
        self._cb_data = cb_data
        self._url = url
        self._target_dir = target_dir

        # May fail with FileNameException, trapped elsewhere
        self._filename = get_base_filename_from_url(self._url, legal_exts)

        if certs and len(certs) > 0:
            self._opener = HTTPSURLopener.HTTPSURLopener(certs, 20)
        else:
            self._opener = urllib.URLopener()
        threading.Thread.__init__(self)

    def run(self):
        result = None
        if self._url and self._target_dir and self._filename:
            if not os.path.exists(self._target_dir):
                os.makedirs(self._target_dir)
            target_file = os.path.join(self._target_dir, self._filename)
            try:
                result = self._opener.retrieve(self._url, target_file)
#            except SSL.SSLError, e:
#                # Don't traceback on dropped connections or timeouts
#                if CommonErrors.canIgnoreSSLError(e):
#                    pass
            except socket.error, e:
                if CommonErrors.canIgnoreSocketError(e):
                    pass
            except IOError, e:
                if CommonErrors.canIgnoreSocketError(e):
                    pass

        if result:
            self._callback('done', self._cb_data)
        else:
            self._callback('failed', self._cb_data)



###########################################################
# Testing stuff
###########################################################

import sys, time

class dlwr:
    def __init__(self, x, t):
        self.num = x
        self.tracker = t
        self.dl = None

    def set_dl(self, dl):
        self.dl = dl


class dl_tracker:
    def __init__(self):
        self.lst = []

    def add(self, dlwr):
        self.lst.append(dlwr)

    def remove(self, dlwr):
        self.lst.remove(dlwr)

    def num(self):
        return len(self.lst)


def dl_callback(status, dlwr):
    print "Finished with %d (%s)" % (dlwr.num, status)
    dlwr.tracker.remove(dlwr)


if __name__ == '__main__':
    if len(sys.argv) < 4:
        print "Usage: python FileDownloader.py key_and_cert ca_cert peer_ca_cert"
        sys.exit(1)

    certs = {}
    certs['key_and_cert'] = sys.argv[1]
    certs['ca_cert'] = sys.argv[2]
    certs['peer_ca_cert'] = sys.argv[3]

    print "Starting..."
    dlt = dl_tracker()

    x = 0
    while x < 100:
        wr = dlwr(x, dlt)
        dstdir = os.path.join("/tmp", "client_dir", "%s" % x)
        if not os.path.exists(dstdir):
            os.makedirs(dstdir)
        time.sleep(0.25)
        dl = FileDownloader(dl_callback, wr, "https://localhost:8886/testfile.dat", dstdir, ['.dat'], certs)
        dlt.add(wr)
        dl.start()
        x = x + 1

    while dlt.num() > 0:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            print "Quitting..."
            os._exit(0)
